@@ -1,17 +1,21 @@
 #!/usr/bin/env python3
 
 import os
+import sys
 import json
 import math
 import uuid
 import click
 import requests
 import dateparser
+import pandas as pd
+
 from datetime import datetime
 from collections.abc import MutableMapping
 from urllib.parse import urlparse, urlunparse
+from centrifuge_cli.policy import CentrifugePolicyCheck
+from centrifuge_cli import __version__ as PACKAGE_VERSION
 
-import pandas as pd
 
 pd.set_option('display.max_rows', None)
 pd.set_option('display.max_columns', None)
@@ -38,6 +42,7 @@ class Cli(object):
         self.limit = limit
         self.outfmt = outfmt
         self.fields = fields
+        self.echo_enabled = True
 
         url = urlparse(endpoint)
         self.endpoint_scheme = url.scheme
@@ -45,8 +50,11 @@ class Cli(object):
 
         self.ssl_verify = not(ssl_no_verify)
 
-    def build_url(self, path, query_list):
-        default_query = [f'limit={self.limit}',
+    def build_url(self, path, query_list, limit):
+        # vulnerable-files does not properly support the limit parameter :facepalm:
+        if ('vulnerable-files' in path):
+            limit = 100
+        default_query = [f'limit={limit}',
                          f'authtoken={self.apikey}']
         if query_list is not None:
             default_query.extend(query_list)
@@ -54,37 +62,51 @@ class Cli(object):
         query = '&'.join(default_query)
         return urlunparse((self.endpoint_scheme, self.endpoint_netloc, path, None, query, None))
 
-    def do_GET(self, path, query_list=None, get_all=False):
-
+    def do_GET(self, path, query_list=None, paginated=False):
         # handle paginated queries
-        if get_all:
+        if paginated:
             results = []
             total = 0
             page = 1
             base_query_list = query_list if query_list else []
             while True:
                 updated_query_list = base_query_list + [f'page={page}']
-                url = self.build_url(path, updated_query_list)
-                res = requests.get(url, verify=self.ssl_verify)
-                res.raise_for_status()
+                url = self.build_url(path, updated_query_list, 100)
+                try:
+                    res = requests.get(url, verify=self.ssl_verify)
+                    res.raise_for_status()
+                except requests.exceptions.HTTPError as err:
+                    if res.status_code == 403:
+                        self.echo(f'Access Denied (url: {url})', err=True)
+                        sys.exit(-1)
+                    else:
+                        raise SystemExit(err)
 
                 data = res.json()
-                # handle binary hardness specifically while that API endpoint is not compliant with the rest
+                # handle binary hardness specifically while that API endpoint is not compliant with the rest :facepalm:
                 data = data['checkSecs'] if 'checkSecs' in data else data
                 results.extend(data['results'])
                 count = data['count']
-                total += self.limit
+                total += len(data['results'])
                 page += 1
 
-                if total >= count:
+                if total >= count or (self.limit > -1 and total >= self.limit):
                     break
 
-            return json.dumps({'count': count, 'results': results}, indent=2, sort_keys=True)
+            return json.dumps({'count': total, 'results': results}, indent=2, sort_keys=True)
 
         else:
-            url = self.build_url(path, query_list)
-            res = requests.get(url, verify=self.ssl_verify)
-            res.raise_for_status()
+            url = self.build_url(path, query_list, self.limit)
+            try:
+                res = requests.get(url, verify=self.ssl_verify)
+                res.raise_for_status()
+            except requests.exceptions.HTTPError as err:
+                if res.status_code == 403:
+                    self.echo(f'Access Denied (url: {url})', err=True)
+                    sys.exit(-1)
+                else:
+                    raise SystemExit(err)
+
             data = res.json()
             data = data['checkSecs'] if 'checkSecs' in data else data
 
@@ -110,29 +132,50 @@ class Cli(object):
         return df
 
     def do_POST(self, path, data, files=None, query_list=None):
-        url = self.build_url(path, query_list)
+        url = self.build_url(path, query_list, self.limit)
+        try:
+            res = requests.post(url, data=data, files=files)
+            res.raise_for_status()
+        except requests.exceptions.HTTPError as err:
+            if res.status_code == 403:
+                self.echo(f'Access Denied (url: {url})', err=True)
+                sys.exit(-1)
+            else:
+                raise SystemExit(err)
 
-        res = requests.post(url, data=data, files=files, verify=self.ssl_verify)
-        res.raise_for_status()
         return res
 
     def do_PUT(self, path, data, query_list=None):
-        url = self.build_url(path, query_list)
+        url = self.build_url(path, query_list, self.limit)
+        try:
+            res = requests.put(url, data=data)
+            res.raise_for_status()
+        except requests.exceptions.HTTPError as err:
+            if res.status_code == 403:
+                self.echo(f'Access Denied (url: {url})', err=True)
+                sys.exit(-1)
+            else:
+                raise SystemExit(err)
 
-        res = requests.put(url, data=data, verify=self.ssl_verify)
-        res.raise_for_status()
         return res
 
     def do_DELETE(self, path, query_list=None):
-        url = self.build_url(path, query_list)
+        url = self.build_url(path, query_list, self.limit)
+        try:
+            res = requests.delete(url)
+            res.raise_for_status()
+        except requests.exceptions.HTTPError as err:
+            if res.status_code == 403:
+                self.echo(f'Access Denied (url: {url})', err=True)
+                sys.exit(-1)
+            else:
+                raise SystemExit(err)
 
-        res = requests.delete(url, verify=self.ssl_verify)
-        res.raise_for_status()
+        return('Deleted')
 
-        if res.status_code not in (200, 204):
-            return('Error occurred, could not delete')
-        else:
-            return('Deleted')
+    def echo(self, message, err=False):
+        if self.echo_enabled or err is True:
+            click.echo(message, err=err)
 
 
 pass_cli = click.make_pass_decorator(Cli)
@@ -147,7 +190,7 @@ pass_cli = click.make_pass_decorator(Cli)
 @click.option('--outfmt', default='human', help='Output format of command', type=click.Choice(['human', 'json', 'csv']))
 @click.option('--field', '-f', multiple=True, metavar='FIELD', help="Select field(s) when output is human or csv")
 @click.option('--ssl-no-verify', help="Disables SSL certificate verification", is_flag=True)
-@click.version_option('0.1')
+@click.version_option(PACKAGE_VERSION)
 @click.pass_context
 def cli(ctx, endpoint, apikey, limit, outfmt, field, ssl_no_verify):
     ctx.obj = Cli(endpoint, apikey, limit, outfmt, field, ssl_no_verify)
@@ -162,19 +205,23 @@ def reports(cli):
 @reports.command(name="list")
 @pass_cli
 def list_command(cli):
-    click.echo(cli.do_GET('/api/upload', query_list=['sorters[0][field]=id',
-                                                     'sorters[0][dir]=desc']))
+    result = cli.do_GET('/api/upload', query_list=['sorters[0][field]=id',
+                                                   'sorters[0][dir]=desc'])
+    cli.echo(result)
+    return(result)
 
 
 @reports.command()
 @click.argument('searchterm', required=True)
 @pass_cli
 def search(cli, searchterm):
-    click.echo(cli.do_GET('/api/upload', query_list=['sorters[0][field]=id',
-                                                     'sorters[0][dir]=desc',
-                                                     'filters[0][field]=search',
-                                                     'filters[0][type]=like',
-                                                     f'filters[0][value]={searchterm}']))
+    result = cli.do_GET('/api/upload', query_list=['sorters[0][field]=id',
+                                                   'sorters[0][dir]=desc',
+                                                   'filters[0][field]=search',
+                                                   'filters[0][type]=like',
+                                                   f'filters[0][value]={searchterm}'])
+    cli.echo(result)
+    return(result)
 
 
 @cli.group()
@@ -187,47 +234,62 @@ def report(cli, ufid):
 @report.command()
 @pass_cli
 def delete(cli):
-    click.echo(cli.do_DELETE('/api/upload', query_list=[f'ufid={cli.ufid}', ]))
+    result = cli.do_DELETE('/api/upload', query_list=[f'ufid={cli.ufid}', ])
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def info(cli):
-    click.echo(cli.do_GET(f'/api/upload/details/{cli.ufid}'))
+    result = cli.do_GET(f'/api/upload/details/{cli.ufid}')
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def crypto(cli):
-    click.echo(cli.do_GET(f'/api/report/crypto/{cli.ufid}', query_list=['sorters[0][field]=path',
-                                                                        'sorters[0][dir]=asc']))
+    click.echo('crypto is a deprecated endpoint and will be removed in future releases, please migrate to certificates and privatekeys',
+               err=True)
+    result = cli.do_GET(f'/api/report/crypto/{cli.ufid}', query_list=['sorters[0][field]=path',
+                                                                      'sorters[0][dir]=asc'])
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def passhash(cli):
-    click.echo(cli.do_GET(f'/api/report/passwordhash/{cli.ufid}'))
+    result = cli.do_GET(f'/api/report/passwordhash/{cli.ufid}')
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def guardian(cli):
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/analyzer-results', query_list=['affected=true&sorters[0][field]=name',
-                                                                                  'sorters[0][dir]=asc']))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/analyzer-results', query_list=['affected=true&sorters[0][field]=name',
+                                                                                'sorters[0][dir]=asc'])
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def sbom(cli):
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/components/pathmatches'))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/components/pathmatches')
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='code-summary')
 @pass_cli
 def code_summary(cli):
-    cli.limit = 100
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/vulnerable-files', get_all=True,
-                          query_list=['sorters[0][field]=id', 'sorters[0][dir]=asc']))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/vulnerable-files',
+                        query_list=['sorters[0][field]=id', 'sorters[0][dir]=asc'], paginated=True)
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='code-static')
@@ -239,7 +301,9 @@ def code_static(cli, exid, path):
     if exid and path:
         query_list.append(f'path={path}')
 
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/vulnerable-files/{exid}', query_list=query_list))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/vulnerable-files/{exid}', query_list=query_list)
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='code-emulated')
@@ -251,31 +315,76 @@ def code_emulated(cli, exid, path):
     if exid and path:
         query_list.append(f'path={path}')
 
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/emulated-files/{exid}', query_list=query_list))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/emulated-files/{exid}', query_list=query_list)
+    cli.echo(result)
+    return(result)
 
 
 @report.command()
 @pass_cli
 def certificates(cli):
-    click.echo(cli.do_GET(f'/api/report/crypto/{cli.ufid}/certificates', get_all=True))
+    result = cli.do_GET(f'/api/report/crypto/{cli.ufid}/certificates', paginated=True)
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='private-keys')
 @pass_cli
-def privatekeys(cli):
-    click.echo(cli.do_GET(f'/api/report/crypto/{cli.ufid}/privateKeys', get_all=True))
+def private_keys(cli):
+    result = cli.do_GET(f'/api/report/crypto/{cli.ufid}/privateKeys', paginated=True)
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='public-keys')
 @pass_cli
 def public_keys(cli):
-    click.echo(cli.do_GET(f'/api/report/crypto/{cli.ufid}/publicKeys', get_all=True))
+    result = cli.do_GET(f'/api/report/crypto/{cli.ufid}/publicKeys', paginated=True)
+    cli.echo(result)
+    return(result)
 
 
 @report.command(name='binary-hardening')
 @pass_cli
 def binary_hardening(cli):
-    click.echo(cli.do_GET(f'/api/report/{cli.ufid}/binary-hardening', get_all=True))
+    result = cli.do_GET(f'/api/report/{cli.ufid}/binary-hardening', paginated=True)
+    cli.echo(result)
+    return(result)
+
+
+@report.command(name='check-policy')
+@click.option('--policy-yaml', metavar='FILE', type=click.Path(), help='Centrifuge policy yaml file.', required=True)
+@click.pass_context
+@pass_cli
+def check_policy(cli, ctx, policy_yaml):
+    outfmt = cli.outfmt
+    cli.outfmt = 'json'
+    cli.echo_enabled = False
+    cli.limit = -1
+    certificates_json = json.loads(ctx.invoke(certificates))
+    private_keys_json = json.loads(ctx.invoke(private_keys))
+    binary_hardening_json = json.loads(ctx.invoke(binary_hardening))
+    guardian_json = json.loads(ctx.invoke(guardian))
+    code_summary_json = json.loads(ctx.invoke(code_summary))
+    passhash_json = json.loads(ctx.invoke(passhash))
+
+    policy_obj = CentrifugePolicyCheck(certificates_json,
+                                       private_keys_json,
+                                       binary_hardening_json,
+                                       guardian_json,
+                                       code_summary_json,
+                                       passhash_json)
+
+    policy_obj.check_rules(policy_yaml)
+
+    result = policy_obj.generate_csv()
+    if outfmt == 'json':
+        result = policy_obj.generate_json()
+
+    cli.echo_enabled = True
+    cli.outfmt = outfmt
+    cli.echo(result)
+    return(result)
 
 
 @cli.command()
@@ -283,7 +392,7 @@ def binary_hardening(cli):
 @click.option('--model', metavar='MODEL', help='Model Number', required=True)
 @click.option('--version', metavar='VERSION', help='Firmware Version Number', required=True)
 @click.option('--chunksize', default=2000000, help='Chunk size in bytes to split the file up into when uploading. Default: 2MB')
-@click.argument('filename', required=True)
+@click.argument('filename', type=click.Path(), required=True)
 @pass_cli
 def upload(cli, make, model, version, chunksize, filename):
     with open(filename, 'rb') as upload_file:
@@ -313,7 +422,14 @@ def upload(cli, make, model, version, chunksize, filename):
             }
             res = cli.do_POST('/api/upload/chunky', data, files=files)
         ufid = res.json()['ufid']
-        click.echo(f"Upload complete. Report id is {ufid}")
+        result = f'Upload complete. Report id is {ufid}'
+        if cli.outfmt == 'json':
+            result = res.text
+        elif cli.outfmt == 'csv':
+            result = f'id,\n{ufid}\n'
+
+        cli.echo(result)
+        return(result)
 
 
 @cli.group()
@@ -325,7 +441,9 @@ def users(cli):
 @users.command(name="list")
 @pass_cli
 def user_list(cli):
-    click.echo(cli.do_GET('/api/user'))
+    result = cli.do_GET('/api/user')
+    cli.echo(result)
+    return(result)
 
 
 @users.command()
@@ -359,7 +477,9 @@ def new(cli, email, password, orgid, admin, expires, no_expire):
         'isPermanent': isPermanent,
         'expiresAt': expiresAt}
 
-    click.echo(cli.do_POST('/api/user', post_data))
+    result = cli.do_POST('/api/user', post_data)
+    cli.echo(result)
+    return(result)
 
 
 @cli.group()
@@ -372,7 +492,9 @@ def user(cli, userid):
 @user.command()
 @pass_cli
 def delete(cli):
-    click.echo(cli.do_DELETE(f'/api/user/{cli.userid}'))
+    result = cli.do_DELETE(f'/api/user/{cli.userid}')
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='set-expiration')
@@ -389,7 +511,9 @@ def set_expiration(cli, expires):
         'isPermanent': False,
         'expiresAt': expiresAt}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='set-password')
@@ -399,7 +523,9 @@ def set_password(cli, password):
     put_data = {
         'password': password}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='set-organization-id')
@@ -409,7 +535,9 @@ def set_organization_id(cli, orgid):
     put_data = {
         'organizationId': int(orgid)}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='set-email')
@@ -419,7 +547,9 @@ def set_email(cli, email):
     put_data = {
         'username': email}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='make-permanent')
@@ -429,7 +559,9 @@ def make_permanent(cli):
         'isPermanent': True,
         'expiresAt': "-"}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @user.command(name='make-admin')
@@ -438,7 +570,9 @@ def make_admin(cli):
     put_data = {
         'isAdmin': True}
 
-    click.echo(cli.do_PUT(f'/api/user/{cli.userid}', put_data))
+    result = cli.do_PUT(f'/api/user/{cli.userid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 @cli.group()
@@ -450,7 +584,9 @@ def orgs(cli):
 @orgs.command(name="list")
 @pass_cli
 def orgs_list(cli):
-    click.echo(cli.do_GET('/api/organization'))
+    result = cli.do_GET('/api/organization')
+    cli.echo(result)
+    return(result)
 
 
 @orgs.command()
@@ -461,7 +597,9 @@ def new(cli, ownerid, name):
     post_data = {
         'ownerId': ownerid,
         'name': name}
-    click.echo(cli.do_POST('/api/organization', post_data))
+    result = cli.do_POST('/api/organization', post_data)
+    cli.echo(result)
+    return(result)
 
 
 @cli.group()
@@ -480,7 +618,9 @@ def change(cli, ownerid, name):
         'name': name,
         'ownerId': int(ownerid)}
 
-    click.echo(cli.do_PUT(f'/api/organization/{cli.orgid}', put_data))
+    result = cli.do_PUT(f'/api/organization/{cli.orgid}', put_data)
+    cli.echo(result)
+    return(result)
 
 
 if __name__ == '__main__':
